import logging
from http import HTTPStatus
from time import sleep

from sherpa_client.api.jobs import get_job
from sherpa_client.api.projects import get_project_info
from sherpa_client.api.segments import search_segments
from sherpa_client.api.services import get_services
from sherpa_client.client import SherpaClient
from sherpa_client.models import (
    Credentials,
    SherpaJobBeanStatus,
    SherpaJobBean,
    SegmentHits,
)
from tenacity import retry, stop_after_delay, stop_after_attempt, wait_fixed

logger = logging.getLogger("kt-client")


class KTClient:
    def __init__(self, sherpa_url, sherpa_user, sherpa_pwd):
        self.user = sherpa_user
        self.pwd = sherpa_pwd
        self.url = sherpa_url
        self.connect()

    def connect(self):
        self.sherpa_client = SherpaClient(
            base_url=f"{self.url}/api", verify_ssl=False, timeout=100
        )
        self.sherpa_client.login_with_cookie(
            Credentials(email=self.user, password=self.pwd)
        )

    @retry(stop=(stop_after_delay(120) | stop_after_attempt(10)), wait=wait_fixed(10))
    def check_resource(self):
        r = get_services.sync_detailed(client=self.sherpa_client)
        if not r.is_success:
            if r.status_code == HTTPStatus.UNAUTHORIZED:
                self.connect()
            r.raise_for_status()
        return r.is_success

    @retry(stop=(stop_after_delay(60) | stop_after_attempt(10)), wait=wait_fixed(1))
    def check_project(self, project):
        r = get_project_info.sync_detailed(project, client=self.sherpa_client)
        if not r.is_success:
            if r.status_code == HTTPStatus.UNAUTHORIZED:
                self.connect()
            r.raise_for_status()
        return r.is_success

    def get_segments(self, project, query, limit, simple):
        result = None
        r = search_segments.sync_detailed(
            project,
            client=self.sherpa_client,
            query=query,
            size=limit,
            simple_query=simple
        )
        if r.is_success:
            hits: SegmentHits = r.parsed
            result = hits.hits
        else:
            logger.warning(
                f"Error when searching segments {query} in {project}",
                exc_info=True,
            )
        return result

    @retry(stop=(stop_after_delay(120) | stop_after_attempt(2)), wait=wait_fixed(5))
    def wait_for_completion(self, job_bean: SherpaJobBean, ŵait_time=5):
        if job_bean:
            while job_bean.status not in [
                SherpaJobBeanStatus.COMPLETED,
                SherpaJobBeanStatus.CANCELLED,
                SherpaJobBeanStatus.FAILED,
            ]:
                sleep(ŵait_time)
                r = get_job.sync_detailed(
                    job_bean.project, job_bean.id, client=self.sherpa_client
                )
                if r.is_success:
                    job_bean = r.parsed
                else:
                    if r.status_code == HTTPStatus.UNAUTHORIZED:
                        self.connect()
                    r.raise_for_status()
        return job_bean

    @staticmethod
    def is_success(job_bean):
        return job_bean and job_bean.status == SherpaJobBeanStatus.COMPLETED



import logging
import os
import re
from functools import lru_cache
from typing import List, cast, Type

from pydantic import Field, BaseModel
from pymultirole_plugins.v1.processor import ProcessorParameters, ProcessorBase
from pymultirole_plugins.v1.schema import Document, Sentence, AltText
from sherpa_client.models import SegmentHit

from pyprocessors_search_segments.kt_client import KTClient

logger = logging.getLogger("pymultirole")
SHERPA_URL = os.getenv("SHERPA_URL", "https://sherpa-afp.kairntech.com")
SHERPA_USER = os.getenv("SHERPA_USER")
SHERPA_PWD = os.getenv("SHERPA_PWD")


class SearchSegmentsParameters(ProcessorParameters):
    question_altText: str = Field(
        "question",
        description="""The alternative text where is stored the original question.""",
    )
    project_name: str = Field(
        None,
        description="""Find a way to inject the project name.""",
    )
    limit: int = Field(
        5,
        description="use the limit argument to only fetch a given number of segments.",
    )
    simple_query: bool = Field(
        True,
        description="use the simple query syntax that is more user-friendly. If set to false, the full lucene syntax will be used, but syntax error could occurs (for advanced users only)",
    )
    index: str = Field(
        None,
        description="""Metadata to use as index for segments.""",
    )


class SearchSegmentsProcessor(ProcessorBase):
    __doc__ = """Replace text of the input document by the similar segments."""

    def process(
            self, documents: List[Document], parameters: ProcessorParameters
    ) -> List[Document]:
        # supported_languages = comma_separated_to_list(SUPPORTED_LANGUAGES)
        params: SearchSegmentsParameters = cast(SearchSegmentsParameters, parameters)
        kt_client = get_kt_client()
        if not kt_client.check_project(params.project_name):
            raise ValueError(f"The project {params.project_name} cannot be found")
        try:
            if not kt_client.check_resource():
                logger.error(f"Sherpa server {SHERPA_URL} is not accessible")
            for document in documents:
                altTexts = document.altTexts or []
                altTexts.append(
                    AltText(name=params.question_altText, text=document.text)
                )
                sentences = []
                text = ""
                hits: List[SegmentHit] = kt_client.get_segments(params.project_name, document.text, params.limit, params.simple_query)
                if hits:
                    for i, hit in enumerate(hits):
                        seg = hit.segment
                        htext = seg.text
                        htext = re.sub(r"\s+", " ", htext)
                        metadata = seg.metadata.additional_properties
                        metadata['documentIdentifier'] = seg.document_identifier
                        metadata['documentTitle'] = seg.document_title
                        index = metadata.get(params.index, None)
                        if index is None:
                            stext = f"{i + 1}. {htext}"
                        else:
                            stext = f"{index} - {htext}"
                        sstart = len(text)
                        text += stext
                        send = len(text)
                        sentences.append(
                            Sentence(
                                start=sstart,
                                end=send,
                                metadata=metadata,
                            )
                        )
                        text += "\n\n"

                document.sentences = sentences
                document.metadata = None
                document.altTexts = altTexts
                document.sentences = sentences
                document.text = text
                document.annotations = None
                document.categories = None
        except BaseException as err:
            raise err
        return documents

    @classmethod
    def get_model(cls) -> Type[BaseModel]:
        return SearchSegmentsParameters


@lru_cache(maxsize=None)
def get_kt_client():
    kt_client = KTClient(SHERPA_URL, SHERPA_USER, SHERPA_PWD)
    return kt_client
